import numpy as np
from qutip import * 

def get_jumps(r,cops=2):
    tbjumps = [None]*r.ntraj
    for i in range(r.ntraj):
        i_bjumps = np.where(r.col_which[i]==cops)[0]
        tbjumps[i] = r.col_times[i][i_bjumps]
    return tbjumps


def get_out(tbjumps, tlist):
    ntraj = len(tbjumps)
    tspace  = np.linspace(tlist.min(), tlist.max(),2500)
    
    out = [None]*ntraj
    for select in range(ntraj):
        tjump    = tbjumps[select]
        nums     = len(tjump)
        out[select] = np.array([np.heaviside(tspace-tjump[i],1) for i in range(nums)]).sum(0) if nums>0 else np.zeros(tspace.size)
    out = np.int32(np.array(out))
    return tspace, out
        

def get_stats(out):
    return out.mean(0), out.std(0)


def sel(x,xlist):
    return np.argmin(np.abs(xlist-x))

def indexof(x,xlist,eps):
    return np.where((xlist<=x+eps)*(xlist>=x-eps))[0]


def get_ps(t,tlist, oute, outo, outemax, outomax):
    selt = sel(t, tlist)
    selp = True if outemax[selt]>outomax[selt] else False
    
    if selp: 
        bins=np.arange(oute[:,selt].max()+2)-0.5
        
        ce,n = np.histogram(oute[:,selt], bins=bins) 
        n = n[:-1]+0.5
        pe = np.asarray(ce,dtype='float')/ce.sum()
        
        co,_ = np.histogram(outo[:,selt], bins=bins) 
        po = np.asarray(co,dtype='float')/co.sum()

    else:
        bins=np.arange(outo[:,selt].max()+2)-0.5
        
        co,n = np.histogram(outo[:,selt], bins=bins) 
        n = n[:-1]+0.5
        po = np.asarray(co,dtype='float')/co.sum()
        
        ce,_ = np.histogram(oute[:,selt], bins=bins) 
        pe = np.asarray(ce,dtype='float')/ce.sum()
    return n, pe, po


def get_p(oute_t, outo_t):
    outemax = oute_t.max()
    outomax = outo_t.max()
    selp = True if outemax>outomax else False
    
    if selp: 
        bins=np.arange(outemax+2)-0.5
        
        ce,n = np.histogram(oute_t, bins=bins) 
        n = n[:-1]+0.5
        pe = np.asarray(ce,dtype='float')/ce.sum()
        
        co,_ = np.histogram(outo_t, bins=bins) 
        po = np.asarray(co,dtype='float')/co.sum()

    else:
        bins=np.arange(outomax+2)-0.5
        
        co,n = np.histogram(outo_t, bins=bins) 
        n = n[:-1]+0.5
        po = np.asarray(co,dtype='float')/co.sum()
        
        ce,_ = np.histogram(oute_t, bins=bins) 
        pe = np.asarray(ce,dtype='float')/ce.sum()
    return n, pe, po


def get_fidelities(tspace, oute, outo):
    error = np.zeros(tspace.size)

    outemax = oute.max(0)
    outomax = outo.max(0)
    for select in range(tspace.size):
        n, pe, po = get_ps(tspace[select], tspace, oute, outo, outemax, outomax)        
        Pe = np.cumsum(pe)
        Po = np.cumsum(po)
        ibar = np.argmin(Pe-Po)        
        error[select] = (1-(Po[ibar]-Pe[ibar]))/2
    return 1-error


def get_fidelity(oute_t, outo_t):
    n, pe, po = get_p(oute_t, outo_t)
    Pe = np.cumsum(pe)
    Po = np.cumsum(po)
    ibar = np.argmin(Pe-Po)        
    error = (1-(Po[ibar]-Pe[ibar]))/2
    return 1-error



def get_rejection_mashgrid(t, tspace, oute, outo):
    outemax, outomax = oute.max(0), outo.max(0)
    n, pe, po = get_ps(t, tspace, oute, outo, outemax, outomax)
    Pe, Po = np.cumsum(pe), np.cumsum(po)
    
    nmax    = int(n.max())
    collect = []
    for nbar1 in range(1,nmax):
        for nbar2 in range(nbar1+1, nmax+1):
            error = ((1-Po[nbar2]) + Pe[nbar1])/2 
            fid_corrected  = 1-error
            rejection_rate = ((Po[nbar2]-Po[nbar1]) + (Pe[nbar2]-Pe[nbar1]))/2
            collect.append([nbar1, nbar2, rejection_rate*100, (1-fid_corrected)*100])
    #    print('\rCompleted %d/%d'%(nbar1,nmax-1),end=' ')
    
    nbar1, nbar2, rejection_mg, inf_mg = list(np.array(collect).transpose())
    return nbar1, nbar2, rejection_mg, inf_mg


def extract_InfvsRej(rej_list, nbar1, nbar2, rejection_mg, inf_mg):
    rej_inf = []
    rej_pairs = []
    for rej in rej_list:
        i = indexof(rej, rejection_mg, 0.05*rej)
        rej_inf.append(inf_mg[i].min())
        rej_pairs.append((nbar1[i][np.argmin(inf_mg[i])], nbar2[i][np.argmin(inf_mg[i])]))
    return np.array(rej_inf), rej_pairs

