import numpy as np
from qutip import * 
from scipy.optimize import minimize_scalar
from scipy.interpolate import interp1d

import sys 
import getpass
s = '/home/'+getpass.getuser()
#s = '/Users/'+getpass.getuser()
sys.path.insert(1, '%s/dptqm/Structural codes/Kerr resonator'%s)
from kerr_cavity import *


def sel_list(x,xlist):
    return np.argmin(np.abs(xlist-x))


def get_Ncut(rho, pillow, tresh):
    return pillow+1+int(np.argwhere( np.invert( np.all(np.abs(rho.full())<tresh, axis=0) ) )[-1])


def cubic_root(a,b,c,d):
    Δ0 = b**2-3*a*c + 0j
    Δ1 = 2*b**3 - 9*a*b*c + 27*a**2*d + 0j
    dt = np.sqrt(Δ1**2-4*Δ0**3)
    C  = ( (Δ1 - dt)/2 )**(1/3)
    
    x0 = -(b + C + Δ0/C)/(3*a)
    
    ξ  = (-1+1j*np.sqrt(3))/2
    x1 = -(b + ξ*C + Δ0/C/ξ)/(3*a)
    x2 = -(b + C*ξ**2 + Δ0/C/ξ**2)/(3*a)
    return np.array([x0,x1,x2])


def get_hysarea(Δr, Ur, Fr, γr, returnbounds=False):
    sem = np.zeros((Δr.size,3),dtype='complex')

    for i,Δ in enumerate(Δr):
        poly = [Ur**2, -(2*Ur*Δ) , (γr**2/4 + Δ**2), -Fr**2]
        sem[i,:] = np.round(cubic_root(*poly),10)
        
    cut0 = np.where(sem[:,0].imag!=0)[0]
    cut1 = np.where(sem[:,1].imag!=0)[0]
    cut2 = np.where(sem[:,2].imag!=0)[0]
    
    n0 = np.copy(sem[:,0].real)
    n1 = np.copy(sem[:,1].real)
    n2 = np.copy(sem[:,2].real)
    
    n0[cut0] = None
    n1[cut1] = None
    n2[cut2] = None
    
    h0 = Δr[cut0[-1]]
    try:
        h1 = Δr[cut1[0]]
    except:
        h1 = Δr[-1]
    
    if returnbounds:
        return n0,n1,n2,h0,h1
    else:
        return n0,n1,n2


def get_visibility(res0, res1, gr):
    t  = res0.times
    y0 = res0.expect[0]    
    interp = interp1d(res1.times, res1.expect[0], kind='linear')
    y1 = interp(t)
    
    N0 = gr*np.trapz(y0, t)
    N1 = gr*np.trapz(y1, t)
    dN = (N0-N1)
    return dN


def get_bounds(Δr, nr, n0, n1): 
    try:
        select = np.invert(np.isnan(n1))
        x   = Δr[select]
        ysc = n1[select]
        yfq = nr[select]
        
        select = x<x[np.argmin(np.gradient(ysc,x))]
        x   = x[select]
        ysc = ysc[select]
        yfq = yfq[select]
        
        diff = np.abs(ysc-yfq)/np.array([ysc,yfq]).max(0)
        x0 = x[sel_list(5e-2,diff)]
        
        select = np.invert(np.isnan(n0))
        x   = Δr[select]
        ysc = n0[select]
        yfq = nr[select]
        
        diff = np.abs(ysc-yfq)/np.array([ysc,yfq]).max(0)
        x1 = x[sel_list(1e-1,diff)]
    except:
        x0 = x1 = 0
    return x0, x1
    
    
def estimate_quantumcp(Ur, Fr, γr, deltamax=300, nmax=10000):
    Δ = np.linspace(0,deltamax,nmax)
    n0,n1,n2,h0,h1 = get_hysarea(Δ, Ur, Fr, γr, returnbounds=True)
    
    select = np.invert(np.isnan(n1))
    x  = Δ[select]
    y  = n1[select]
    x0 = x[np.argmin(np.gradient(y,x))]
    
    keep = False if (np.gradient(n1)[0]<0 and np.gradient(n1)[1]<0) else True 
    return x0,h0,h1,keep


def minim0(Δ, N, δ, Ur, Fr, gr):
    rho0  = fock(N,0)
    tlist = np.linspace(0,5,100)
    
    p = Parameters(delta=Δ, U=Ur, G=0, F=Fr, gamma=gr, eta=0)
    k = Kerr(Ncut=N, p=p)
    res0 = mesolve(k.H, rho0, tlist, k.c_ops, [k.a.dag()*k.a])
    
    p = Parameters(delta=Δ+δ, U=Ur, G=0, F=Fr, gamma=gr, eta=0)
    k = Kerr(Ncut=N, p=p)
    res1 = mesolve(k.H, rho0, tlist, k.c_ops, [k.a.dag()*k.a])
    return -get_visibility(res0, res1, gr)


def minim1(Δ, N, generateH, δ, Ur, Fr, gr):
    nsteps, Δt, jp, tmax = 50, 0.05, 0.1, 5
    rho0  = fock(N,0)
    p = Parameters(delta=Δ, U=Ur, Grat=0, F=Fr, gamma=gr, eta=0)    
    res0 = dynamic_mesolve(generateH, (p,), rho0, nsteps, Δt, tmax, jp, pillow=4, tresh=1e-4, fixdt=True)
    
    p = Parameters(delta=Δ+δ, U=Ur, Grat=0, F=Fr, gamma=gr, eta=0)
    res1 = dynamic_mesolve(generateH, (p,), rho0, nsteps, Δt, tmax, jp, pillow=4, tresh=1e-4, fixdt=True)
    return -get_visibility(res0, res1, gr)